/*
 * Copyright (c) 1989, 1993, 1994
 * The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Michael Fischbein.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * This product includes software developed by the University of
 * California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

/*
 * written by shige <shige@kuis.kyoto-u.ac.jp>
 */

#include "jconv.h"
#include <stdio.h>

static unsigned char kanji1st = 0;
static unsigned char kanji2nd = 0;
unsigned char esc_set1byte[2] = {'(', 'B'};
unsigned char esc_set2byte[2] = {'$', 'B'};

/* convert JIS code to Shift-JIS code */
static void jistojms(unsigned char c1, unsigned char c2)
{
	if (c1 & 1) {
		c1 = (c1 >> 1) + 0x71;
		c2 += 0x1f;
		if (c2 >= 0x7f)
			c2++;
		} else {
			c1 = (c1 >> 1) + 0x70;
			c2 += 0x7e;
	}
	if (c1 > 0x9f)
		c1 += 0x40;

	kanji1st = c1;
	kanji2nd = c2;
	return;
}

/* convert Shift-JIS code to JIS code */
static void jmstojis(unsigned char c1, unsigned char c2)
{
	c1 -= (c1 <= 0x9f) ? 0x70 : 0xb0;
	c1 <<= 1;
	if (c2 < 0x9f) {
		c2 -= (c2 < 0x7f) ? 0x1f : 0x20;
		c1--;
	}
	else {
		c2 -= 0x7e;
	}

	kanji1st = c1;
	kanji2nd = c2;
	return;
}

/* Shift-JIS to EUC */
void sjistoeuc(char *s)
{
	unsigned char c1, c2;
	char *sp = s;

	while (*s != '\0') {
		c1 = (unsigned char)(*s);
		if (c1 < 0x80) {
			*s++ = (char)c1;
		} else if (iskanji1st(c1)) {
			c2 = (unsigned char)(*(s+1));
			if (*(s+1) == '\0') {
				*s++ = (char)c1;
				break;
			}
			if (iskanji2nd(c2)) {
				jmstojis(c1, c2);
				*s++ = (char)(kanji1st | 0x80);
				*s++ = (char)(kanji2nd | 0x80);
			} else {
				*s++ = (char)c1;
				*s++ = (char)c2;
			}
		} else if (iskana(c1)) {
			*s++ = 0x8e;
			*s++ = (char)c1;
		} else {
			*s++ = (char)c1;
		}
	}
	*s = '\0';

	return;
}

/* Shift-JIS to JIS */
void sjistojis(char *s)
{
	unsigned char c1, c2, c3;
	char state, kana;
	char _7bit;
	char *t;

	t = (char *)malloc(2 * strlen(s));
	strcpy(t, s);
	/* _7bit = _7bitmode; */
	_7bit = 1;
	state = kana = 0;

	c1 = (unsigned char)(*t++);
	while (*t != '\0') {
		if (c1 < 0x80) {
			if (state) {
				/* escape sequence for 1 byte character */ 
				*s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
				state = 0;
			}
			else if (kana) {
				*s++ = SI;
				kana = 0;
			}
			*s++ = c1;
		} else if (iskanji1st(c1)) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				/* JIS end */
				if (kana) {
					*s++ = SI;
				} else if (state) {
					/* JIS escape sequence for 1 byte character */
					*s++ = ESC;
					*s++ = esc_set1byte[0];
					if (esc_set1byte[1])
						*s++ = esc_set1byte[1];
				}
				*s++ = c1;
				break;
			}
			if (!state) {
				if (kana) {
					*s++ = SI;
					kana = 0;
				}
				/* escape sequence for 2 byte character */ 
				*s++ = ESC;
				*s++ = esc_set2byte[0];
				if (esc_set2byte[1])
					*s++ = esc_set2byte[1];
				state = 1;
			}
			if (iskanji2nd(c2)) {
				jmstojis(c1, c2);
				*s++ = kanji1st;
				*s++ = kanji2nd;
			} else {
				*s++ = c1;
				*s++ = c2;
			}
		} else if (iskana(c1)) {
			if (state) {
				/* escape sequence for 1 byte character */ 
				*s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
				state = 0;
			}
			if (_7bit) {
				if (!kana) {
					*s++ = SO;
					kana = 1;
				}
				c1 &= 0x7f;
			}
			*s++ = c1;
		} else {
			if (state) {
				/* escape sequence for 1 byte character */ 
				*s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
				state = 0;
			} else if (kana) {
				*s++ = SI;
				kana = 0;
			}
			*s++ = c1;
		}

		/* next chars */
		c1 = (unsigned char)(*t++);
		if (c1 == '\0') {
			/* JIS end */
			if (kana) {
				*s++ = SI;
			} else if (state) {
				/* JIS escape sequence for 1 byte character */
				*s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
			}
			break;
		}
	}
	*s = '\0';

	return;
}

/* EUC to Shift-JIS */
void euctosjis(char *s)
{
	unsigned char c1, c2;
	char *t;

	t = (char *)malloc(2 * strlen(s));
	strcpy(t, s);

	while (*t != '\0') {
		c1 = (unsigned char)(*t++);
		if (c1 < 0x80) {
			*s++ = (char)c1;
		} else if (iseuc(c1)) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				*s++ = (char)c1;
				break;
			}
			if (iseuc(c2)) {
				jistojms(c1 & 0x7f, c2 & 0x7f);
				*s++ = (char)kanji1st;
				*s++ = (char)kanji2nd;
			} else {
				*s++ = (char)c1;
				*s++ = (char)c2;
			}
		} else if (c1 == 0x8e) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0')
				break;
			*s++ = (char)c2;
		} else if (c1 == 0x8f) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0')
				break;
			*s++ = (char)c2;
			c2 = (unsigned char)(*t++);
			if (c2 == '\0')
				break;
			*s++ = (char)c2;
        	} else {
			*s++ = (char)c1;
		}
	}
	*s = '\0';

	return;
}

/* EUC to JIS */
void euctojis(char *s)
{
	unsigned char c1, c2, c3;
	char state, kana;
	char _7bit;
	char *t;

	t = (char *)malloc(2 * strlen(s));
	strcpy(t, s);
	/* _7bit = _7bitmode; */
	_7bit = 1;
	state = kana = 0;

	c1 = (unsigned char)(*t++);
	while (*t != '\0') {
		if (c1 < 0x80) {
			if (state) {
				/* JIS escape sequence for 1 byte character */
                		*s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
				state = 0;
			}
			else if (kana) {
                		*s++ = SI;
				kana = 0;
			}
			*s++ = c1;
		} else if (iseuc(c1)) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				/* JIS end */
				if (kana) {
					*s++ = SI;
				} else if (state) {
					/* JIS escape sequence for 1 byte character */
	                		*s++ = ESC;
					*s++ = esc_set1byte[0];
					if (esc_set1byte[1])
						*s++ = esc_set1byte[1];
				}

				*s++ = c1;
				break;
			}
			if (!state) {
				if (kana) {
                			*s++ = SI;
					kana = 0;
				}
				/* JIS escape sequence for 2 byte character */
                		*s++ = ESC;
				*s++ = esc_set2byte[0];
				if (esc_set2byte[1])
					*s++ = esc_set2byte[1];
				state = 1;
			}
			if (iseuc(c2)) {
				*s++ = (char)(c1 & 0x7f);
				*s++ = (char)(c2 & 0x7f);
			} else {
				*s++ = (char)c1;
				*s++ = (char)c2;
			}
		} else if (c1 == 0x8e) {
			c1 = (unsigned char)(*t++);
			if (c1 == '\0') {
				/* JIS end */
				if (kana) {
					*s++ = SI;
				} else if (state) {
					/* JIS escape sequence for 1 byte character */
	                		*s++ = ESC;
					*s++ = esc_set1byte[0];
					if (esc_set1byte[1])
						*s++ = esc_set1byte[1];
				}
				break;
			}
			if (state) {
				/* JIS escape sequence for 1 byte character */
	                	*s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
				state = 0;
			}
			if (_7bit) {
				if (!kana) {
					*s++ = SO;
					kana = 1;
				}
				c1 &= 0x7f;
			}
			*s++ = c1;
		} else if (c1 == 0x8f) {
			c1 = (unsigned char)(*t++);
			if (c1 == '\0') {
				/* JIS end */
				if (kana) {
					*s++ = SI;
				} else if (state) {
					/* JIS escape sequence for 1 byte character */
		                	*s++ = ESC;
					*s++ = esc_set1byte[0];
					if (esc_set1byte[1])
						*s++ = esc_set1byte[1];
				}
				break;
			}
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				/* JIS end */
				if (kana) {
					*s++ = SI;
				} else if (state) {
					/* JIS escape sequence for 1 byte character */
		                	*s++ = ESC;
					*s++ = esc_set1byte[0];
					if (esc_set1byte[1])
						*s++ = esc_set1byte[1];
				}
				*s++ = c1;
				break;
			}
			if (!state) {
				if (kana) {
					*s++ = SI;
					kana = 0;
				}
				/* JIS escape sequence for 2 byte character */
                		*s++ = ESC;
				*s++ = esc_set2byte[0];
				if (esc_set2byte[1])
					*s++ = esc_set2byte[1];
				state = 1;
			}
			*s++ = c1;
			*s++ = c2;
		} else {
			if (state) {
				/* JIS escape sequence for 1 byte character */
		                *s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
				state = 0;
			} else if (kana) {
				*s++ = SI;
				kana = 0;
			}
			*s++ = c1;
		}

		/* next chars */
		c1 = (unsigned char)(*t++);
		if (c1 == '\0') {
			/* JIS end */
			if (kana) {
				*s++ = SI;
			} else if (state) {
				/* JIS escape sequence for 1 byte character */
		                *s++ = ESC;
				*s++ = esc_set1byte[0];
				if (esc_set1byte[1])
					*s++ = esc_set1byte[1];
			}
			break;
		}
	}
	*s = '\0';

	return;
}

/* JIS to EUC */
void jistoeuc(char *s)
{
	unsigned char c1, c2, c3;
	char state, kana;
	char _7bit;
	char *t;

	t = (char *)malloc(2 * strlen(s));
	strcpy(t, s);
	/* _7bit = _7bitmode; */
	_7bit = 1;
	state = kana = 0;

	c1 = (unsigned char)(*t++);
	while (*t != '\0') {
		if (c1 == ESC) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				*s++ = (char)c1;
				break;
			}
			switch (c2) {
			case '$':
				c3 = (unsigned char)(*t++);
				if (c3 == '\0') {
					*s++ = (char)c1;
					*s++ = (char)c2;
					break;
				}
				if (c3 == 'B' || c3 == '@')
					state = 1;
				else {
					*s++ = (char)c1;
					*s++ = (char)c2;
					c1 = c3;
					continue;
				}
				break;
			case '(':
				c3 = (unsigned char)(*t++);
				if (c3 == '\0') {
					*s++ = (char)c1;
					*s++ = (char)c2;
					break;
				}
				if (c3 == 'J' || c3 == 'B' || c3 == 'H')
					state = 0;
				else {
					*s++ = (char)c1;
					*s++ = (char)c2;
					c1 = c3;
					continue;
				}
				break;
			case 'K':
				state = 1;
				break;
			case 'H':
				state = 0;
				break;
			default:
				*s++ = (char)c1;
				continue;
			}
		} else if (c1 <= 0x20 || c1 == 0x7f) {
			if (_7bit && (c1 == SO || c1 == SI))
				kana = c1 == SO;
			else
				*s++ = (char)c1;
		} else if (state) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				*s++ = (char)c1;
				break;
			}
			if (c2 <= 0x20 || c2 == 0x7f) {
				*s++ = (char)c1;
				c1 = c2;
				continue;
			}
			*s++ = (char)(c1 | 0x80);
			*s++ = (char)(c2 | 0x80);
		} else {
			if (_7bit && kana)
				c1 |= 0x80;
			if (iskana(c1))
				*s++ = (char)0x8e;
			*s++ = (char)c1;
		}

		/* next chars */
		c1 = (unsigned char)(*t++);
		if (c1 == '\0') {
			break;
		}
	}
	*s = '\0';

	return;
}

/* JIS to Shift-JIS */
void jistosjis(char *s)
{
	unsigned char c1, c2, c3;
	char state, kana;
	char _7bit;
	char *t;

	t = (char *)malloc(2 * strlen(s));
	strcpy(t, s);
	/* _7bit = _7bitmode; */
	_7bit = 1;
	state = kana = 0;

	c1 = (unsigned char)(*t++);
	while (*t != '\0') {
		if (c1 == ESC) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				*s++ = (char)c1;
				break;
			}
			switch (c2) {
			case '$':
				c3 = (unsigned char)(*t++);
				if (c3 == '\0') {
					*s++ = (char)c1;
					*s++ = (char)c2;
					break;
				}
				if (c3 == 'B' || c3 == '@')
					state = 1;
				else {
					*s++ = (char)c1;
					*s++ = (char)c2;
					c1 = c3;
					continue;
				}
				break;
			case '(':
				c3 = (unsigned char)(*t++);
				if (c3 == '\0') {
					*s++ = (char)c1;
					*s++ = (char)c2;
					break;
				}
				if (c3 == 'J' || c3 == 'B' || c3 == 'H')
					state = 0;
				else {
					*s++ = (char)c1;
					*s++ = (char)c2;
					c1 = c3;
					continue;
				}
				break;
			case 'K':
				state = 1;
				break;
			case 'H':
				state = 0;
				break;
			default:
				*s++ = (char)c1;
				continue;
			}
		} else if (c1 <= 0x20 || c1 == 0x7f) {
			if (_7bit && (c1 == SO || c1 == SI))
				kana = c1 == SO;
			else
				*s++ = (char)c1;
		} else if (state) {
			c2 = (unsigned char)(*t++);
			if (c2 == '\0') {
				*s++ = (char)c1;
				break;
			}
			if (c2 <= 0x20) {
				*s++ = (char)c1;
				c1 = c2;
				continue;
			}
			if (c1 < 0x80 && isjis(c2)) {
				jistojms(c1, c2);
				*s++ = (char)kanji1st;
				*s++ = (char)kanji2nd;
			} else {
				*s++ = (char)c1;
				*s++ = (char)c2;
			}
		} else {
			if (_7bit && kana)
				c1 |= 0x80;
			*s++ = (char)c1;
		}

		/* next chars */
		c1 = (unsigned char)(*t++);
		if (c1 == '\0') {
			break;
		}
	}
	*s = '\0';

	return;
}

/* scan forward if Shift-JIS code or JIS code exists */
int checkforward(p, size)
unsigned char *p;
unsigned int size;
{
    unsigned char c;

    if (size > 1024)
        size = 1024;
    if (size == 0)
        return EUC;
    c = *p++;
    while (1) {
        if (c == ESC) {
            if (--size == 0)
                break;
            if ((c = *p++) == '$') {
                if (--size == 0)
                    break;
                if ((c = *p++) == 'B' || c == '@')
                    return JIS;
                else
                    continue;
            }
            else if (c == 'K')
                return JIS;
            else
                continue;
        }
        else if (c >= 0x81) {
            if (c == 0x8e) {
                if (--size == 0)
                    break;
                p++;
            }
            else if (c <= 0x9f) {
                if (--size == 0)
                    break;
                c = *p++;
                if (iskanji2nd(c))
                    return SJIS;
                else
                    continue;
            }
            else if (c >= 0xa1 && c <= 0xdf) {
                if (--size == 0)
                    break;
                c = *p++;
                if (iskana(c))
                    continue;
                else if (iseuc(c))
                    return EUC;
                else
                    continue;
            }
            else if (c != 0xa0)
                return EUC;
        }
        if (--size == 0)
            break;
        c = *p++;
    }
    return EUC;
}

/* detect kanji code */
int detectcode(buf, size)
unsigned char *buf;
unsigned int size;
{
    register unsigned char *p, c;
    register unsigned int count;
    int unknownstat;

    count = size;
    if (count == 0)
        return ASCII;
    p = buf;
    unknownstat = 0;
    c = *p++;
    while (1) {
        if (c == ESC) {
            if (--count == 0)
                break;
            if ((c = *p++) == '$') {
                if (--count == 0)
                    break;
                if ((c = *p++) == 'B' || c == '@')
                    return JIS;
                else
                    continue;
            }
            else if (c == 'K')
                return JIS;
            else
                continue;
        }
        else if (c >= 0x81) {
            if (c == 0x8e) {
                if (--count == 0)
                    break;
                c = *p++;
                if (iskana(c))
                    unknownstat |= 1;
                else if (iskanji2nd(c))
                    return SJIS;
                else
                    continue;
            }
            else if (c <= 0x9f) {
                if (--count == 0)
                    break;
                c = *p++;
                if (iskanji2nd(c))
                    return SJIS;
                else
                    continue;
            }
            else if (c >= 0xa1 && c <= 0xdf || c == 0xfd || c == 0xfe) {
                if (--count == 0)
                    break;
                c = *p++;
                if (iseuc(c))
                    if (iskana(c))
                        return checkforward(p, count - 1);
                    else
                        return EUC;
                else
                    continue;
            }
            else if (c >= 0xe0 && c <= 0xfc) {
                if (--count == 0)
                    break;
                c = *p++;
                if (iskanji2nd(c))
                    if (iseuc(c))
                        unknownstat |= 1;
                    else
                        return SJIS;
                else
                    if (iseuc(c))
                        return EUC;
                    else
                        continue;
            }
        }
        if (--count == 0)
            break;
        c = *p++;
    }
    switch (unknownstat) {
    case 1:
    case 3:
        return UNKNOWN;
    case 2:
        return SJIS;
    default:
        return ASCII;
    }
}
